"use strict";
/*
 * Copyright Fastly, Inc.
 * Licensed under the MIT license. See LICENSE file for details.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.doTransform = exports.validateConfig = void 0;
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const constants_1 = require("next/constants");
const constants_2 = require("./constants");
const util_1 = require("./util");
const file_1 = require("./file");
const DIST_DIR = '.next'; // vercel build always uses .next
const NEXT_LAUNCHER_FILENAME = '___next_launcher.cjs';
const FASTLY_NEXT_LAUNCHER_FILENAME = '___fastly_next_launcher.cjs';
const INIT_SCRIPT_FILENAME = `next-compute-js-server-${constants_2.NEXT_VERSION}.cjs`;
const NEXT_FILES_ROOT_DEST = `/lib/next-${constants_2.NEXT_VERSION}`;
function validateConfig(vcConfig) {
    if (vcConfig.handler !== NEXT_LAUNCHER_FILENAME) {
        throw new Error(`.vc-config.json contains unexpected 'launcher' value. Expected '${NEXT_LAUNCHER_FILENAME}', found '${vcConfig.handler ?? '(no value)'}'.`);
    }
}
exports.validateConfig = validateConfig;
function doTransform(vcConfig, ctx) {
    // Create /lib/ directory if it doesn't already exist
    const libDir = path_1.default.join(ctx.buildOutputPath, 'lib');
    if (!fs_1.default.existsSync(libDir)) {
        fs_1.default.mkdirSync(libDir, { recursive: true });
    }
    // Copy /lib/next-NEXT_VERSION/.next files.  This needs to happen only once regardless of the
    // number of times this plugin runs.
    const nextRootDest = path_1.default.join(ctx.buildOutputPath, NEXT_FILES_ROOT_DEST);
    if (!fs_1.default.existsSync(nextRootDest)) {
        const srcDir = path_1.default.join(ctx.nextProjectPath, DIST_DIR);
        (0, file_1.copyFiles)(srcDir, path_1.default.join(nextRootDest, DIST_DIR), `'${DIST_DIR}' files`, (src) => {
            if (src.endsWith('.js.map') ||
                src.endsWith('.nft.json') ||
                src.startsWith(path_1.default.join(srcDir, 'cache')) ||
                src.startsWith(path_1.default.join(srcDir, 'static')) ||
                src.startsWith(path_1.default.join(srcDir, 'trace'))) {
                console.log(`Skipping ${src}`);
                return false;
            }
            return true;
        });
    }
    // Create /init/ script. This needs to happen only once regardless of the number of
    // times this plugin runs.
    const initScriptPath = path_1.default.join(ctx.buildOutputPath, 'init', INIT_SCRIPT_FILENAME);
    if (!fs_1.default.existsSync(initScriptPath)) {
        fs_1.default.mkdirSync(path_1.default.dirname(initScriptPath), { recursive: true });
        const initScript = buildInitScript(ctx.transformName);
        (0, file_1.writeFile)(initScriptPath, initScript, 'Plugin Init script');
    }
    console.log(`Next.js nodejs transform STARTING for '${ctx.functionPath}'.`);
    // Create the output directory
    fs_1.default.mkdirSync(ctx.functionFilesTargetPath, { recursive: true });
    // Next.js config taken from SERVER_FILES_MANIFEST from project root
    const nextConfig = loadNextConfig(ctx.nextProjectPath);
    const conf = Object.assign({}, nextConfig, { compress: false });
    // Generate and write the fastly launcher script
    const fastlyNextLauncherScript = buildFastlyNextLauncherScript({
        nextRuntimePackage: ctx.transformName,
        conf,
    });
    (0, file_1.writeFile)(path_1.default.join(ctx.functionFilesTargetPath, FASTLY_NEXT_LAUNCHER_FILENAME), fastlyNextLauncherScript, 'Fastly Next.js Launcher');
    // Create a new .vc-config.json file for the transformed function
    const newVcConfig = {
        runtime: 'edge',
        deploymentTarget: 'v8-worker',
        name: (0, util_1.mapFunctionPathToFunctionName)(ctx.functionPath),
        entrypoint: FASTLY_NEXT_LAUNCHER_FILENAME,
        framework: {
            slug: 'nextjs',
            version: constants_2.NEXT_VERSION,
        },
    };
    (0, file_1.writeFile)(path_1.default.join(ctx.functionFilesTargetPath, constants_2.VERCEL_FUNCTION_CONFIG_FILENAME), JSON.stringify(newVcConfig, undefined, 2), 'Function Config file');
    console.log(`Next.js nodejs transform COMPLETED for '${ctx.functionPath}'.`);
}
exports.doTransform = doTransform;
function loadNextConfig(projectPath) {
    const nextDirectory = path_1.default.join(projectPath, DIST_DIR);
    const serverFilesManifestPath = path_1.default.join(nextDirectory, constants_1.SERVER_FILES_MANIFEST);
    let serverFilesManifest;
    try {
        const serverFilesManifestJson = fs_1.default.readFileSync(serverFilesManifestPath, 'utf-8');
        serverFilesManifest = JSON.parse(serverFilesManifestJson);
    }
    catch (ex) {
        throw new Error(`Error loading server files manifest file \`${serverFilesManifestPath}'.`, { cause: ex });
    }
    if (serverFilesManifest.version !== 1 || serverFilesManifest.config == null) {
        throw new Error(`Server files manifest file \`${serverFilesManifestPath}' missing or invalid`);
    }
    // If the JSON was successfully read, we assume it's in that format.
    return serverFilesManifest.config;
}
function buildFastlyNextLauncherScript(data) {
    return `
    const { default: NextComputeJsServer, initFs } = require(${JSON.stringify(data.nextRuntimePackage)});
    
    const conf = ${JSON.stringify(data.conf)};
    
    initFs("${path_1.default.join(NEXT_FILES_ROOT_DEST)}");
    
    const nextServer = new NextComputeJsServer({
      conf,
      computeJsConfig: {
        extendRenderOpts: {
          runtime: "experimental-edge",
        },
      },
      minimalMode: true,
      customServer: false,
    });
    const handler = nextServer.getComputeJsRequestHandler();

    module.exports = async (request, context) => {
      return await handler(request);
    };
  `;
}
function buildInitScript(nextRuntimePackage) {
    return `
    module.exports = function({contentAssets, moduleAssets}) {
      const { init } = require(${JSON.stringify(nextRuntimePackage)});
      init({contentAssets, moduleAssets});
    };
  `;
}
