/*
 * Copyright Fastly, Inc.
 * Licensed under the MIT license. See LICENSE file for details.
 *
 * Portions of this file Copyright Vercel, Inc., licensed under the MIT license. See LICENSE file for details.
 */
import { join, resolve } from 'path';
import { APP_PATHS_MANIFEST, FONT_MANIFEST, PAGES_MANIFEST, SERVER_DIRECTORY, } from 'next/constants';
import { normalizeLocalePath } from 'next/dist/shared/lib/i18n/normalize-locale-path';
import { denormalizePagePath } from 'next/dist/shared/lib/page-path/denormalize-page-path';
import { normalizePagePath } from 'next/dist/shared/lib/page-path/normalize-page-path';
import { MissingStaticPage, PageNotFoundError } from 'next/dist/shared/lib/utils';
import { getFsSettings } from './fs';
import LRUCache from 'lru-cache';
const pagePathCache = new LRUCache({
    max: 1000,
});
const loadManifest = (manifestPath) => {
    return requireManifest(manifestPath);
};
export function getMaybePagePath(page, distDir, locales, isAppPath) {
    const cacheKey = `${page}:${distDir}:${locales}:${isAppPath}`;
    if (pagePathCache.has(cacheKey)) {
        return pagePathCache.get(cacheKey);
    }
    const serverBuildPath = join(distDir, SERVER_DIRECTORY);
    let appPathsManifest;
    if (isAppPath) {
        appPathsManifest = loadManifest(join(serverBuildPath, APP_PATHS_MANIFEST));
    }
    const pagesManifest = loadManifest(join(serverBuildPath, PAGES_MANIFEST));
    try {
        page = denormalizePagePath(normalizePagePath(page));
    }
    catch (err) {
        console.error(err);
        throw new PageNotFoundError(page);
    }
    const checkManifest = (manifest) => {
        let curPath = manifest[page];
        if (!manifest[curPath] && locales) {
            const manifestNoLocales = {};
            for (const key of Object.keys(manifest)) {
                manifestNoLocales[normalizeLocalePath(key, locales).pathname] =
                    pagesManifest[key];
            }
            curPath = manifestNoLocales[page];
        }
        return curPath;
    };
    let pagePath;
    if (appPathsManifest) {
        pagePath = checkManifest(appPathsManifest);
    }
    if (!pagePath) {
        pagePath = checkManifest(pagesManifest);
    }
    if (!pagePath) {
        pagePathCache.set(cacheKey, null);
        return null;
    }
    const path = join(serverBuildPath, pagePath);
    pagePathCache.set(cacheKey, path);
    return path;
}
/**
 * Finds the path that corresponds to a page, based on the pages manifest and localizations.
 * (An adaptation for Compute of function in Next.js of the same name,
 * found at next/server/require.ts)
 */
export function getPagePath(page, distDir, locales, isAppPath) {
    const pagePath = getMaybePagePath(page, distDir, locales, isAppPath);
    if (!pagePath) {
        throw new PageNotFoundError(page);
    }
    return pagePath;
}
/**
 * Loads the string or module that corresponds to a page.
 * (An adaptation for Compute of function in Next.js of the same name,
 * found at next/server/require.ts)
 */
export async function requirePage(page, distDir, isAppPath) {
    const pagePath = getPagePath(page, distDir, undefined, isAppPath);
    if (pagePath.endsWith('.html')) {
        try {
            return readAssetFileAsString(pagePath);
        }
        catch (err) {
            throw new MissingStaticPage(page, err.message);
        }
    }
    return await requireModule(pagePath);
}
/**
 * Load the font manifest.
 * (An adaptation for Compute of function in Next.js of the same name,
 * found at next/server/require.ts)
 */
export function requireFontManifest(distDir) {
    const serverBuildPath = join(distDir, SERVER_DIRECTORY);
    const fontManifest = requireManifest(join(serverBuildPath, FONT_MANIFEST));
    return fontManifest;
}
/* ---- */
export function assetDirectoryExists(path) {
    const { contentAssets, dir } = getFsSettings();
    const relativePath = resolve(dir, path);
    return contentAssets.getAssetKeys()
        .some(key => key.startsWith(relativePath + '/'));
}
export function assetDirectory(path) {
    const { contentAssets, dir } = getFsSettings();
    const relativePath = resolve(dir, path);
    return contentAssets.getAssetKeys()
        .filter(key => key.startsWith(relativePath + '/'));
}
export function assetFileExists(path) {
    const { contentAssets, dir } = getFsSettings();
    const relativePath = resolve(dir, path);
    return contentAssets.getAsset(relativePath) != null;
}
export function readAssetFileAsString(path) {
    const { contentAssets, dir } = getFsSettings();
    const relativePath = resolve(dir, path);
    const file = contentAssets.getAsset(relativePath);
    return file?.getText() ?? '';
}
export function requireManifest(path) {
    let content = readAssetFileAsString(path) ?? '';
    return JSON.parse(content);
}
export async function requireModule(path) {
    // TODO: eventually move this out into its own package
    const { moduleAssets, dir } = getFsSettings();
    // Do Node.js magic to find file
    const relativePath = resolve(dir, path);
    const file = moduleAssets.getAsset(relativePath) ??
        moduleAssets.getAsset(relativePath + '.js') ??
        // NOTE: Node.js would check .json here, before index.js
        moduleAssets.getAsset(relativePath + '/index.js') ??
        null;
    if (file == null) {
        throw { relativePath, 'code': 'MODULE_NOT_FOUND' };
    }
    return file.getModule();
}
